from abc import ABC, abstractmethod
from decimal import Decimal
from importlib import import_module
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Type, Union

from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ImproperlyConfigured
from django.forms import BaseForm
from django.http import HttpRequest, HttpResponse
from django.views import View

from getpaid.types import ChargeResponse, PaymentStatusResponse

if TYPE_CHECKING:
    from .models import AbstractPayment
else:
    from django.db.models import Model as AbstractPayment


class BaseProcessor(ABC):
    production_url = None  #: Base URL of production environment.
    sandbox_url = None  #: Base URL of sandbox environment.
    display_name = None  #: The name of the provider for the ``choices``.
    accepted_currencies = None  #: List of accepted currency codes (ISO 4217).
    logo_url = None  #: Logo URL - can be used in templates.
    slug = None  #: For friendly urls
    post_form_class = None
    post_template_name = None
    client_class = None
    client = None
    ok_statuses = [
        200,
    ]  #: List of potentially successful HTTP status codes returned by paywall when creating payment

    def __init__(self, payment: AbstractPayment) -> None:
        self.payment = payment
        self.path = payment.backend
        self.context = {}  # can be used by Payment's customized methods.
        if self.slug is None:
            self.slug = self.path
        self.config = getattr(settings, "GETPAID_BACKEND_SETTINGS", {}).get(
            self.path, {}
        )
        self.optional_config = getattr(settings, "GETPAID", {})
        if self.client_class is not None:
            self.client = self.get_client()

    def get_client_class(self) -> Type:
        class_path = self.get_setting("CLIENT_CLASS")
        if not class_path:
            class_path = self.client_class
        if class_path and not callable(class_path):
            module_name, _, class_name = class_path.rpartition(".")
            module = import_module(module_name)
            return getattr(module, class_name)
        return class_path

    def get_client(self) -> object:
        return self.get_client_class()(**self.get_client_params())

    def get_client_params(self) -> dict:
        return {}

    @classmethod
    def class_id(cls, **kwargs) -> str:
        return cls.__module__

    def get_setting(self, name: str, default: Optional[Any] = None) -> Any:
        value = self.config.get(name, default)
        if value is None:
            value = self.optional_config.get(name, None)
        return value

    @classmethod
    def get_display_name(cls, **kwargs) -> str:
        return cls.display_name

    @classmethod
    def get_accepted_currencies(cls, **kwargs) -> List[str]:
        return cls.accepted_currencies

    @classmethod
    def get_logo_url(cls, **kwargs) -> str:
        return cls.logo_url

    @classmethod
    def get_paywall_baseurl(cls, **kwargs) -> str:
        if settings.DEBUG:
            return cls.sandbox_url
        return cls.production_url

    @staticmethod
    def get_our_baseurl(request: HttpRequest = None, **kwargs) -> str:
        """
        Little helper function to get base url for our site.
        Note that this way 'https' is enforced on production environment.
        """
        if request is None:
            return "http://127.0.0.1/"
        scheme = "http" if settings.DEBUG else "https"
        return f"{scheme}://{get_current_site(request).domain}/"

    def get_template_names(self, view: Optional[View] = None, **kwargs) -> List[str]:
        template_name = self.get_setting("POST_TEMPLATE")
        if template_name is None:
            template_name = self.post_template_name
        if template_name is None and hasattr(view, "get_template_names"):
            return view.get_template_names()
        if template_name is None:
            raise ImproperlyConfigured("Couldn't determine template name!")
        return [template_name]

    def get_form_class(self, **kwargs) -> Type:
        form_class_path = self.get_setting("POST_FORM_CLASS")
        if not form_class_path:
            return self.post_form_class
        if isinstance(form_class_path, str):
            module_path, class_name = form_class_path.rsplit(".", 1)
            module = import_module(module_path)
            return getattr(module, class_name)
        return self.post_form_class

    def prepare_form_data(self, post_data: dict, **kwargs) -> Mapping[str, Any]:
        """
        If backend support several modes of operation, POST should probably
        additionally calculate some sort of signature based on passed data.
        """
        return post_data

    def get_form(self, post_data: dict, **kwargs) -> BaseForm:
        """
        (Optional)
        Used to get POST form for backends that use such flow.
        """
        form_class = self.get_form_class()
        if form_class is None:
            raise ImproperlyConfigured("Couldn't determine form class!")

        form_data = self.prepare_form_data(post_data)

        return form_class(fields=form_data)

    # Communication with outer world

    @abstractmethod
    def prepare_transaction(
        self, request: HttpRequest, view: Optional[View] = None, **kwargs
    ) -> HttpResponse:
        """
        Prepare Response for the view asking to prepare transaction.

        :return: HttpResponse instance
        """
        raise NotImplementedError

    def handle_paywall_callback(self, request: HttpRequest, **kwargs) -> HttpResponse:
        """
        This method handles the callback from paywall for the purpose
        of asynchronously updating the payment status in our system.

        :return: HttpResponse instance that will be presented as answer to the callback.
        """
        raise NotImplementedError

    def fetch_payment_status(self, **kwargs) -> PaymentStatusResponse:
        # TODO use interface annotation to specify the dict layout
        """
        Logic for checking payment status with paywall.
        """
        raise NotImplementedError

    def charge(
        self, amount: Optional[Union[Decimal, float, int]] = None, **kwargs
    ) -> ChargeResponse:
        """
        (Optional)
        Check if payment can be locked and call processor's method.
        This method is used eg. in flows that pre-authorize payment during
        order placement and charge money later.
        """
        raise NotImplementedError

    def release_lock(self, **kwargs) -> Decimal:
        """
        (Optional)
        Release locked payment. This can happen if pre-authorized payment cannot
        be fullfilled (eg. the ordered product is no longer available for some reason).
        Returns released amount.
        """
        raise NotImplementedError

    def start_refund(
        self, amount: Optional[Union[Decimal, float, int]] = None, **kwargs
    ) -> Decimal:
        """
        Refunds the given amount.

        Returns the amount that is refunded.
        """
        raise NotImplementedError

    def cancel_refund(self, **kwargs) -> bool:
        """
        Cancels started refund.

        Returns True/False if the cancel succeeded.
        """
        raise NotImplementedError



from django import template

from getpaid.registry import registry

register = template.Library()


@register.simple_tag
def get_backends(currency):
    """
    Get all backends supporting given currency into template's context in raw form.
    This way you can use all fields to render backend chooser.
    """
    return registry.get_backends(currency)



from importlib import import_module

from django.conf import settings


def run_getpaid_validators(data):
    backend = data["backend"]
    getpaid_settings = getattr(settings, "GETPAID", {})
    global_validators = getpaid_settings.get("VALIDATORS", [])
    backend_validators = (
        getpaid_settings.get("BACKENDS", {}).get(backend, {}).get("VALIDATORS", [])
    )
    for path in set(global_validators).union(backend_validators):
        module_name, validator_name = path.rsplit(".", 1)
        module = import_module(module_name)
        validator = getattr(module, validator_name)
        data = validator(data)
    return data



class GetPaidException(Exception):
    def __init__(self, *args, **kwargs):
        self.context = kwargs.pop("context", None)
        super().__init__(*args)


class CommunicationError(GetPaidException):
    pass


class ChargeFailure(CommunicationError):
    pass


class LockFailure(CommunicationError):
    pass


class RefundFailure(CommunicationError):
    pass


class CredentialsError(GetPaidException):
    pass



""""
Settings:
    pos_id
    second_key
    client_id
    client_secret
"""
import json
import logging
import os
from urllib.parse import urljoin

import requests
from django.http import HttpResponse, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import reverse, reverse_lazy
from django_fsm import can_proceed

from getpaid.post_forms import PaymentHiddenInputsPostForm
from getpaid.processor import BaseProcessor
from getpaid.status import PaymentStatus as ps

logger = logging.getLogger(__name__)


class PaymentProcessor(BaseProcessor):
    slug = "dummy"
    display_name = "Dummy"
    accepted_currencies = [
        "PLN",
        "EUR",
    ]
    ok_statuses = [200]
    method = "REST"  # Supported modes: REST, POST, GET
    confirmation_method = "PUSH"  # PUSH or PULL
    post_form_class = PaymentHiddenInputsPostForm
    post_template_name = "dummy/payment_post_form.html"
    _token = None
    standard_url = reverse_lazy("paywall:gateway")
    api_url = reverse_lazy("paywall:api_register")

    def get_paywall_method(self):
        return self.get_setting("paywall_method", self.method)

    def get_confirmation_method(self):
        return self.get_setting("confirmation_method", self.confirmation_method).upper()

    def get_paywall_baseurl(self, request=None, **kwargs):
        if request is None:
            base = os.environ.get("_PAYWALL_URL")
        else:
            base = os.environ["_PAYWALL_URL"] = request.build_absolute_uri("/")
        if self.get_paywall_method() == "REST":
            return urljoin(base, str(self.api_url))
        return urljoin(base, str(self.standard_url))

    def get_params(self):
        base = self.get_paywall_baseurl()
        params = {
            "ext_id": self.payment.id,
            "value": self.payment.amount_required,
            "currency": self.payment.currency,
            "description": self.payment.description,
            "success_url": urljoin(
                base,
                reverse("getpaid:payment-success", kwargs={"pk": str(self.payment.pk)}),
            ),
            "failure_url": urljoin(
                base,
                reverse("getpaid:payment-failure", kwargs={"pk": str(self.payment.pk)}),
            ),
        }
        if self.get_confirmation_method() == "PUSH":
            params["callback"] = urljoin(
                base, reverse("getpaid:callback", kwargs={"pk": str(self.payment.pk)})
            )
        return {k: str(v) for k, v in params.items()}

    # Specifics
    def prepare_transaction(self, request, view=None, **kwargs):
        target_url = self.get_paywall_baseurl(request)
        params = self.get_params()
        method = self.get_paywall_method()
        if method == "REST":
            response = requests.post(target_url, json=params)
            if response.status_code in self.ok_statuses:
                self.payment.confirm_prepared()
                self.payment.save()
            return HttpResponseRedirect(response.json()["url"])
        elif method == "POST":
            self.payment.confirm_prepared()
            self.payment.save()
            form = self.get_form(params)
            return TemplateResponse(
                request=request,
                template=self.get_template_names(view=view),
                context={"form": form, "paywall_url": target_url},
            )
        else:
            # GET payments are a bit tricky. You can either confirm payment as
            # prepared here, or on successful return from paywall.
            self.payment.confirm_prepared()
            self.payment.save()
            return HttpResponseRedirect(target_url)

    def handle_paywall_callback(self, request, **kwargs):
        new_status = json.loads(request.body).get("new_status")
        if new_status is None:
            raise ValueError("Got no status")
        elif new_status == ps.FAILED:
            self.payment.fail()
        elif new_status == ps.PRE_AUTH:
            self.payment.confirm_lock()
        elif new_status == ps.PAID:
            if can_proceed(self.payment.confirm_lock):  # GET flow needs this
                self.payment.confirm_lock()
            if can_proceed(self.payment.confirm_payment):
                self.payment.confirm_payment()
                if can_proceed(self.payment.mark_as_paid):
                    self.payment.mark_as_paid()
                elif can_proceed(self.payment.mark_as_refunded):
                    self.payment.mark_as_refunded()
        else:
            raise ValueError(f"Unhandled new status {new_status}")
        self.payment.save()
        return HttpResponse("OK")

    def fetch_payment_status(self, **kwargs):
        base = self.get_paywall_baseurl()
        response = requests.get(
            urljoin(
                base,
                reverse(
                    "paywall:get_status", kwargs={"pk": str(self.payment.external_id)}
                ),
            )
        )
        if response.status_code not in self.ok_statuses:
            raise Exception("Error occurred!")
        status = response.json()["payment_status"]
        results = {}
        if status == ps.PAID:
            results["callback"] = "confirm_payment"
        elif status == ps.PRE_AUTH:
            results["callback"] = "confirm_lock"
        elif status == ps.PREPARED:
            results["callback"] = "confirm_prepared"
        elif status == ps.FAILED:
            results["callback"] = "fail"
        return results

    def charge(self, amount=None, **kwargs):
        url = urljoin(self.get_paywall_baseurl(), reverse("paywall:api_operate"))
        requests.post(
            url, json={"id": str(self.payment.external_id), "new_status": ps.PAID}
        )

    def release_lock(self, **kwargs):
        url = urljoin(self.get_paywall_baseurl(), reverse("paywall:api_operate"))
        requests.post(
            url, json={"id": str(self.payment.external_id), "new_status": ps.REFUNDED}
        )

    def start_refund(self, amount=None, **kwargs):
        url = urljoin(self.get_paywall_baseurl(), reverse("paywall:api_operate"))
        requests.post(
            url,
            json={
                "id": str(self.payment.external_id),
                "new_status": ps.REFUND_STARTED,
            },
        )

    def cancel_refund(self, **kwargs):
        url = urljoin(self.get_paywall_baseurl(), reverse("paywall:api_operate"))
        requests.post(
            url, json={"id": str(self.payment.external_id), "new_status": ps.PAID}
        )



from django.urls import include, path

from . import views
from .registry import registry

app_name = "getpaid"

urlpatterns = [
    path("new/", views.new_payment, name="create-payment"),
    path("success/<uuid:pk>/", views.success, name="payment-success",),
    path("failure/<uuid:pk>/", views.failure, name="payment-failure",),
    path("callback/<uuid:pk>/", views.callback, name="callback",),
    path("", include(registry.urls)),
]



import importlib

from django.urls import include, path

from getpaid.processor import BaseProcessor


def importable(name):
    try:
        importlib.import_module(name)
        return True
    except ImportError:
        return False


class PluginRegistry(object):
    def __init__(self):
        self._backends = {}

    def __contains__(self, item):
        return item in self._backends

    def __getitem__(self, item):
        return self._backends[item]

    def __iter__(self):
        return iter(self._backends)

    def register(self, module_or_proc):
        """
        Register module containing PaymentProcessor class or a PaymentProcessor directly.
        """
        if hasattr(module_or_proc, "__base__") and issubclass(
            module_or_proc, BaseProcessor
        ):
            self._backends[module_or_proc.slug] = module_or_proc
        else:
            processor = module_or_proc.processor.PaymentProcessor
            self._backends[module_or_proc.__name__] = processor

    def get_choices(self, currency):
        """
        Get CHOICES for plugins that support given currency.
        """
        currency = currency.upper()
        return [
            (name, p.display_name)
            for name, p in self._backends.items()
            if currency in p.get_accepted_currencies()
        ]

    def get_backends(self, currency):
        """
        Get plugins that support given currency.
        """
        currency = currency.upper()
        return [
            backend
            for backend in self._backends.values()
            if currency in backend.get_accepted_currencies()
        ]

    @property
    def urls(self):
        """
        Provide URL structure for all registered plugins that have urls defined.
        """
        return [
            path(
                "{}/".format(p.slug),
                include(("{}.urls".format(name), p.slug), namespace=p.slug),
            )
            for name, p in self._backends.items()
            if importable("{}.urls".format(name))
        ]

    def get_all_supported_currency_choices(self):
        """
        Get all currencies that are supported by at least one plugin,
        in CHOICES format.
        """
        currencies = set()
        for backend in self._backends:
            currencies.update(backend.accepted_currencies or [])
        return [(c.upper(), c.upper()) for c in currencies]


registry = PluginRegistry()



from django.contrib import admin

from . import models


# Payment model is used here directly so that this PaymentAdmin does not show
# if Payment is swapped.
@admin.register(models.Payment)
class PaymentAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "order_id",
        "amount_required",
        "currency",
        "status",
        "backend",
        "created_on",
        "last_payment_on",
        "amount_paid",
    )
    search_fields = ("id", "order_id")
    date_hierarchy = "created_on"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['getpaid',
 'getpaid.backends',
 'getpaid.backends.dummy',
 'getpaid.migrations',
 'getpaid.templatetags']

package_data = \
{'': ['*'], 'getpaid.backends.dummy': ['templates/getpaid_dummy/*']}

install_requires = \
['django-fsm>=2.7.0,<3.0.0',
 'django-model-utils>=4.0.0,<5.0.0',
 'pendulum>=2.0.5,<3.0.0',
 'swapper>=1.1.2,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'django-getpaid',
    'version': '2.2.0',
    'description': 'Multi-broker payment processor for Django.',
    'long_description': '.. image:: https://img.shields.io/pypi/v/django-getpaid.svg\n    :target: https://pypi.org/project/django-getpaid/\n    :alt: Latest PyPI version\n.. image:: https://img.shields.io/travis/sunscrapers/django-getpaid.svg\n    :target: https://travis-ci.org/sunscrapers/django-getpaid\n.. image:: https://api.codacy.com/project/badge/Coverage/d25ba81e2e4740d6aac356f4ac90b16d\n    :target: https://www.codacy.com/manual/dekoza/django-getpaid\n.. image:: https://img.shields.io/pypi/wheel/django-getpaid.svg\n    :target: https://pypi.org/project/django-getpaid/\n.. image:: https://img.shields.io/pypi/l/django-getpaid.svg\n    :target: https://pypi.org/project/django-getpaid/\n.. image:: https://api.codacy.com/project/badge/Grade/d25ba81e2e4740d6aac356f4ac90b16d\n    :target: https://www.codacy.com/manual/dekoza/django-getpaid\n\n=============================\nWelcome to django-getpaid\n=============================\n\n\ndjango-getpaid is payment processing framework for Django\n\nDocumentation\n=============\n\nThe full documentation is at https://django-getpaid.readthedocs.io.\n\nFeatures\n========\n\n* support for multiple payment brokers at the same time\n* very flexible architecture\n* support for asynchronous status updates - both push and pull\n* support for modern REST-based broker APIs\n* support for multiple currencies (but one per payment)\n* support for global and per-plugin validators\n* easy customization with provided base abstract models and swappable mechanic (same as with Django\'s User model)\n\n\nQuickstart\n==========\n\nInstall django-getpaid and at least one payment backend:\n\n.. code-block:: console\n\n    pip install django-getpaid\n    pip install django-getpaid-payu\n\nAdd them to your ``INSTALLED_APPS``:\n\n.. code-block:: python\n\n    INSTALLED_APPS = [\n        ...\n        \'getpaid\',\n        \'getpaid_payu\',  # one of plugins\n        ...\n    ]\n\nAdd getpaid to URL patterns:\n\n.. code-block:: python\n\n    urlpatterns = [\n        ...\n        path(\'payments/\', include(\'getpaid.urls\')),\n        ...\n    ]\n\nDefine an ``Order`` model by subclassing ``getpaid.models.AbstractOrder``\nand define some required methods:\n\n.. code-block:: python\n\n    from getpaid.models import AbstractOrder\n\n    class MyCustomOrder(AbstractOrder):\n        amount = models.DecimalField(decimal_places=2, max_digits=8)\n        description = models.CharField(max_length=128)\n        buyer = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)\n\n        def get_absolute_url(self):\n            return reverse(\'order-detail\', kwargs={"pk": self.pk})\n\n        def get_total_amount(self):\n            return self.amount\n\n        def get_buyer_info(self):\n            return {"email": self.buyer.email}\n\n        def get_currency(self):\n            return "EUR"\n\n        def get_description(self):\n            return self.description\n\n.. note:: If you already have an Order model and don\'t want to subclass ``AbstractOrder``\n    just make sure you implement all methods.\n\nInform getpaid of your Order model in ``settings.py`` and provide settings for payment backends:\n\n.. code-block:: python\n\n    GETPAID_ORDER_MODEL = \'yourapp.MyCustomOrder\'\n\n    GETPAID_BACKEND_SETTINGS = {\n        "getpaid_payu": {\n            # take these from your merchant panel:\n            "pos_id": 12345,\n            "second_key": "91ae651578c5b5aa93f2d38a9be8ce11",\n            "oauth_id": 12345,\n            "oauth_secret": "12f071174cb7eb79d4aac5bc2f07563f",\n        },\n    }\n\nWrite a view that will create the Payment.\n\nAn example view and its hookup to urls.py can look like this:\n\n.. code-block:: python\n\n    # orders/views.py\n    from getpaid.forms import PaymentMethodForm\n\n    class OrderView(DetailView):\n        model = Order\n\n        def get_context_data(self, **kwargs):\n            context = super(OrderView, self).get_context_data(**kwargs)\n            context["payment_form"] = PaymentMethodForm(\n                initial={"order": self.object, "currency": self.object.currency}\n            )\n            return context\n\n    # main urls.py\n\n    urlpatterns = [\n        # ...\n        path("order/<int:pk>/", OrderView.as_view(), name="order_detail"),\n    ]\n\nYou\'ll also need a template (``order_detail.html`` in this case) for this view.\nHere\'s the important part:\n\n.. code-block::\n\n    <h2>Choose payment broker:</h2>\n    <form action="{% url \'getpaid:create-payment\' %}" method="post">\n        {% csrf_token %}\n        {{ payment_form.as_p }}\n        <input type="submit" value="Checkout">\n    </form>\n\n\nRunning Tests\n=============\n\n.. code-block:: console\n\n    poetry install\n    poetry run tox\n\n\nAlternatives\n============\n\n* `django-payments <https://github.com/mirumee/django-payments>`_\n\n\nCredits\n=======\n\nCreated by `Krzysztof Dorosz <https://github.com/cypreess>`_.\nRedesigned and rewritten by `Dominik Kozaczko <https://github.com/dekoza>`_.\n\nProudly sponsored by `SUNSCRAPERS <http://sunscrapers.com/>`_\n\n\n\nDisclaimer\n==========\n\nThis project has nothing in common with `getpaid <http://code.google.com/p/getpaid/>`_ plone project.\n',
    'author': 'Dominik Kozaczko',
    'author_email': 'dominik@kozaczko.info',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/django-getpaid/django-getpaid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from django import forms


class PaymentHiddenInputsPostForm(forms.Form):
    def __init__(self, fields, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for key in fields:
            self.fields[key] = forms.CharField(
                initial=fields[key], widget=forms.HiddenInput
            )
